/*
	FPGA Test Module
	Brief:
		Success: BLINK_GOOD flashes at 1Hz on test board
		Failure:
			1) BLINK_GOOD is off => FPGA and potential MCU failure
			2) BLINK_GOOD is on and not flashing => MCU failure

			
	Pin assignments:
		PIN_88:  clk_in, 40MHz clock
		PIN_111: clk_out, ~1Hz clock output to MCU (PA12)
		PIN_112: mcu_in, ~1Hz echoed clock input from MCU (PA13)
		PIN_84:  mcu_out, BLINK_GOOD output
	
	If the microcontroller has failed, the echoed clock will not appear on mcu_in,
	and mcu_out will remain high
	
	If the FPGA has failed, mcu_out will remain low.
*/

module fpga_testpcb (
							input logic clk_in,
							input logic mcu_in,
							output logic clk_out,
							output logic mcu_out
						  );
	
	assign clk_out = clk_divide[23];
	assign mcu_out = mcu_in;
		
	reg [24:0] clk_divide;
	always @(posedge clk_in) 
	begin
		clk_divide <= clk_divide + 1;
	end
	

endmodule
	
						  
							